; Boot sector to load up an OS kernel
; Based partly on the Linux kernel
; Platform: Intel x86

; Status: Works allright. Could use portability improvements, though, as
; noted in comments. Also, load_file needs a little fleshing out.
; If it loads from a *real* floppy disk, not just in DOSEMU, it's good
; enough for me.

%include "segments.inc"

start:
        ;Move the boot sector to 0x90000
        mov     ax, BOOTSEG
        mov     ds, ax
        mov     ax, INITSEG
        mov     es, ax
        mov     cx, 256
        sub     si, si
        sub     di, di
        cld
        rep
        movsw
        mov     ds, ax  ;DS=ES=INITSEG
        jmp     INITSEG:go

go:
        mov     di, 0x4000-12   ;Setup data & stack segs
        mov     ds, ax
        mov     ss, ax
        mov     sp, di

        mov     bp, BootMsg
        call    print_string

        ;Get disk parameters (by trial & error). Try 36, 18, 15, and 9 sectors
probe
        mov     si, DiskSizes
.loop
        lodsb   ;AL=[SI]
        cbw     ;AX=AL (extended to word length)
        mov     [sectors], ax
        cmp     si, DiskSizes+4
        jae     .done
        xchg    ax, cx
        xor     dx, dx
        xor     bl, bl
        mov     bh, [sectors]
        inc     bh
        shl     bh, 1
        mov     ax, 0x0201
        int     0x13
        jc      .loop
.done
        mov     ax, INITSEG
        mov     es, ax


load:
        ; reset floppy drive
        xor     ah, ah
        xor     dl, dl
        int     13h

        ;print "Loading setup" message
        mov     bp,SetupMsg
        call    print_string

        ;load setup
        mov     ax,SETUPSEG
        mov     cx,[Ssize]
        call    load_file

        ;print "Loading kernel" message
        mov     bp,KernelMsg
        call    print_string

        ;load the kernel
        mov     ax,SYSSEG
        mov     cx,[Ksize]
        call    load_file

        ;Shut off floppy drive
        call    kill_motor

        ;print "Starting" message
        mov     bp,StartMsg
        call    print_string
        call    print_nl

        jmp     SETUPSEG:0      ;Run the kernel



;Parameters:
;       AX = segment to load to
;       CX = number of 512-byte blocks to load
;Register usage:
;       AX,DX = calculations
;       BX = free
;       CX = can't touch
load_file:
        mov     es,ax
        mov     [loadseg],ax

.check:         ;Calculate total # of sectors read so far. Done yet?
        mov     ax,es
        sub     ax,[loadseg]
        shr     ax,5    ;AX = # of sectors read, total
        cmp     ax,cx   ;Finished?
        jb      .calc
        ret
.calc:
        mov     dx,cx
        sub     dx,ax           ;# of sectors left to load
        mov     ax,[sectors]
        sub     ax,[sect]       ;AX = # of sectors remaining on track

        ;if (DX < AX), AX=DX
        cmp     ax,dx
        jbe     .read
        mov     ax,dx           ;AX = # of sectors to load
.read:
        call    read_track

        ;End of track reached?
        add     ax,[sect]       ;AX = # of sectors read on track
        cmp     ax,[sectors]
        jb      .loop
.next:
        xor     ax,ax           ;go to sector 0 of next track
        inc     word [track]

        ;need to increment disk head, if the kernel gets really big someday,
        ;or it's used on an old computer


.loop:
        mov     [sect],ax       ;Update sect

        ;ES += AX << 5
        shl     ax,5            ;# of paragraphs read
        mov     dx,es
        add     ax,dx           ;Update ES
        mov     es,ax

        jmp     .check


;Parameters:
; AL = # of sectors to read
; ES:0 = address to read into
; track/sect/head = where to start from
read_track:
        pusha

        xor     bx,bx   ;Start loading at 1000:0000
        mov     dx,[track]
        mov     cx,[sect]       ;set sector
        inc     cx
        mov     ch,dl           ;set track
        mov     dx,[head]       ;set head
        mov     dh,dl
        and     dx,0100h        ;drive=0; head=0 or 1
        mov     ah,2

        ;For error dump (if necessary)
        push    dx
        push    cx
        push    bx
        push    ax

        int     13h
        jc      bad_rt
        add     sp,8

        ; print a "." (2Eh) for every track loaded, like Linux
        mov     ax,0E2Eh
        mov     bx,0007h
        int     10h

        popa
        ret

bad_rt:
        ;debug/error message
        push    ax
        call    print_all

        ;reset drive
        xor     ah,ah
        xor     dl,dl
        int     13h

;We should give up after too many tries

        ;retry
        popa
        jmp     hang
        jmp     read_track


debug:
        pusha
        push    dx
        push    cx
        push    bx
        push    ax
        push    ax
        call    print_all
        popa
        ret

; Register debugging dump (borrowed from Linux bootloader)
; Expects this on stack: dx, cx, bx, ax, errorcode
print_all:
        mov     cx,5    ;1 error code + 4 registers
        mov     bp,sp

print_loop:
        push    cx      ;save CX reg
        call    print_nl

        cmp     cl,5    ;if (cx < 5) skip the next part
        jae     no_reg

        mov     ax,0E05h + 'A' - 1
        sub     al,cl
        int     10h
        mov     al,'X'
        int     10h
        mov     al,':'
        int     10h

no_reg:
        add     bp,2
        call    print_hex
        pop     cx
        loop    print_loop
        ret

; Supporting print functions
print_nl:
        mov     ax,0E0Dh        ;Service 0E=TTY character output
        int     10h
        mov     al,0Ah
        int     10h
        ret

print_hex:      ;prints word at (ss:bp)
        mov     cx,4
        mov     dx,[bp]
print_digit:
        rol     dx,4
        mov     ax,0E0Fh
        and     al,dl
        add     al,90h
        daa
        adc     al,40h
        daa
        int     10h
        loop    print_digit
        ret

;expects ES:BP to point to a null-terminated string
;
print_string:
        mov     ah,0Eh
        mov     bx,0            ;Set video page & Text olor
.l1     cmp     [ds:bp],byte 0
        je      .end
        mov     al,[ds:bp]
        int     10h
        inc     bp
        jmp     .l1
.end    ret

;Turns off the floppy drive motor
kill_motor:
        push    dx
        mov     dx,03F2h
        xor     al,al
        out     dx,al
        pop     dx
        ret

hang:
        jmp hang


BootMsg         db "Boot loader",13,10,0
BootMsgLen      equ $-BootMsg

SetupMsg        db "Loading Setup",0
SetupMsgLen     equ $-SetupMsg

KernelMsg       db 13,10,"Loading Kernel",0
KernelMsgLen    equ $-KernelMsg

StartMsg        db 13,10,"Starting",0
StartMsgLen     equ $-StartMsg

;Disk specs
sectors:        dw      0

;Variables for load_file
sect:   dw 1            ;current sector
track:  dw 0            ;current track
head:   dw 0            ;current head
loadseg: dw 0           ;Destination segment

DiskSizes       db 36,18,15,9


TIMES   506-($-$$) DB 0         ; skip to end of sector
;Used by the DOS loader, etc.
Ssize   DW      SETUPSIZE
Ksize   DW      SYSSIZE

DW      0xAA55                  ; signature
